from abc import ABC, abstractmethod
from typing import Dict, List


class FinalQuery:
    def __init__(self, query):
        self._query = query

    def __repr__(self):
        return f'{self.__class__.__name__}("{self._query}")'


class Query:
    def __init__(self, query: str, columns: List[str]):
        self.query = query
        self.columns = columns

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.query}", columns={self.columns})'

    def select(self, *columns: str):
        if not (set(columns) <= set(self.columns)):
            raise ValueError(
                f"Unknown column names passed: {set(columns) - set(self.columns)}"
            )

        selected_columns = ", ".join(columns)
        return FinalQuery(f"SELECT {selected_columns} FROM {self.query}")

    def select_star(self):
        return FinalQuery(f"SELECT * FROM {self.query}")

    def where(self, where_clause: str):
        return Query(
            f"{self.query} WHERE {where_clause}",
            columns=self.columns,
        )

    def insert_into(self) -> FinalQuery:
        columns = ", ".join(self.columns)
        placeholders = ", ".join(["?"] * len(self.columns))
        return FinalQuery(
            f"INSERT INTO {self.query}({columns}) VALUES ({placeholders})"
        )

    def drop_table(self) -> FinalQuery:
        return FinalQuery(f"DROP TABLE {self.query}")

    def delete_from(self) -> FinalQuery:
        return FinalQuery(f"DELETE FROM {self.query}")


class Model(ABC):
    @property
    @abstractmethod
    def table_name(self) -> str:
        ...

    @property
    @abstractmethod
    def schema(self) -> Dict[str, str]:
        ...

    @classmethod
    def create_table(cls, connection):
        metadata_table = [
            f"{column_name} {column_type}"
            for column_name, column_type in cls.schema.items()
        ]
        metadata_table_definition = ", ".join(metadata_table)
        connection.execute(
            f"CREATE TABLE IF NOT EXISTS {cls.table_name}({metadata_table_definition})"
        )


class Name(Model):
    table_name = "names"
    schema = {
        "name": "TEXT",
        "module": "TEXT",
        "package": "TEXT",
        "source": "INTEGER",
        "type": "INTEGER",
    }
    columns = list(schema.keys())
    objects = Query(table_name, columns)

    @classmethod
    def create_table(cls, connection):
        super().create_table(connection)
        connection.execute("CREATE INDEX IF NOT EXISTS name ON names(name)")
        connection.execute("CREATE INDEX IF NOT EXISTS module ON names(module)")
        connection.execute("CREATE INDEX IF NOT EXISTS package ON names(package)")

    search_submodule_like = objects.where('module LIKE ("%." || ?)')
    search_module_like = objects.where("module LIKE (?)")

    import_assist = objects.where("name LIKE (? || '%')")

    search_by_name_like = objects.where("name LIKE (?)")

    delete_by_module_name = objects.where("module = ?").delete_from()


class Package(Model):
    table_name = "packages"
    schema = {
        "package": "TEXT",
        "path": "TEXT",
    }
    columns = list(schema.keys())
    objects = Query(table_name, columns)

    delete_by_package_name = objects.where("package = ?").delete_from()



import rope.base.builtins  # Use full qualification for clarity.
from rope.base import ast, codeanalyze, exceptions, pynames, utils
from rope.refactor import patchedast


class Scope:
    def __init__(self, pycore, pyobject, parent_scope):
        self.pycore = pycore
        self.pyobject = pyobject
        self.parent = parent_scope

    def get_names(self):
        """Return the names defined or imported in this scope"""
        return self.pyobject.get_attributes()

    def get_defined_names(self):
        """Return the names defined in this scope"""
        return self.pyobject._get_structural_attributes()

    def get_name(self, name):
        """Return name `PyName` defined in this scope"""
        if name not in self.get_names():
            raise exceptions.NameNotFoundError("name %s not found" % name)
        return self.get_names()[name]

    def __getitem__(self, key):
        """The same as ``get_name(key)``"""
        return self.get_name(key)

    def __contains__(self, key):
        """The same as ``key in self.get_names()``"""
        return key in self.get_names()

    @utils.saveit
    def get_scopes(self):
        """Return the subscopes of this scope

        The returned scopes should be sorted by the order they appear.
        """
        return self._create_scopes()

    def lookup(self, name):
        if name in self.get_names():
            return self.get_names()[name]
        if self.parent is not None:
            return self.parent._propagated_lookup(name)
        return None

    def get_propagated_names(self):
        """Return the visible names of this scope

        Return the names defined in this scope that are visible from
        scopes containing this scope.  This method returns the same
        dictionary returned by `get_names()` except for `ClassScope`
        which returns an empty dict.
        """
        return self.get_names()

    def _propagated_lookup(self, name):
        if name in self.get_propagated_names():
            return self.get_propagated_names()[name]
        if self.parent is not None:
            return self.parent._propagated_lookup(name)
        return None

    def _create_scopes(self):
        return [
            pydefined.get_scope() for pydefined in self.pyobject._get_defined_objects()
        ]

    def _get_global_scope(self):
        current = self
        while current.parent is not None:
            current = current.parent
        return current

    def get_start(self):
        return self.pyobject.get_ast().lineno

    def get_body_start(self):
        body = self.pyobject.get_ast().body
        if body:
            return body[0].lineno
        return self.get_start()

    def get_end(self):
        pymodule = self._get_global_scope().pyobject
        return pymodule.logical_lines.logical_line_in(self.logical_end)[1]

    @utils.saveit
    def get_logical_end(self):
        global_scope = self._get_global_scope()
        return global_scope._scope_finder.find_scope_end(self)

    start = property(get_start)
    end = property(get_end)
    logical_end = property(get_logical_end)

    def get_kind(self):
        pass

    def get_region(self):
        self._calculate_scope_regions_for_module()
        node = self.pyobject.get_ast()
        region = patchedast.node_region(node)
        return region

    def _calculate_scope_regions_for_module(self):
        self._get_global_scope()._calculate_scope_regions()

    def in_region(self, offset):
        """Checks if offset is in scope region"""

        region = self.get_region()
        return region[0] < offset < region[1]


class GlobalScope(Scope):
    def __init__(self, pycore, module):
        super().__init__(pycore, module, None)
        self.names = module._get_concluded_data()

    def get_start(self):
        return 1

    def get_kind(self):
        return "Module"

    def get_name(self, name):
        try:
            return self.pyobject[name]
        except exceptions.AttributeNotFoundError:
            if name in self.builtin_names:
                return self.builtin_names[name]
            raise exceptions.NameNotFoundError("name %s not found" % name)

    @utils.saveit
    def _calculate_scope_regions(self):
        source = self._get_source()
        patchedast.patch_ast(self.pyobject.get_ast(), source)

    def _get_source(self):
        return self.pyobject.source_code

    def get_names(self):
        if self.names.get() is None:
            result = dict(self.builtin_names)
            result.update(super().get_names())
            self.names.set(result)
        return self.names.get()

    def get_inner_scope_for_line(self, lineno, indents=None):
        return self._scope_finder.get_holding_scope(self, lineno, indents)

    def get_inner_scope_for_offset(self, offset):
        return self._scope_finder.get_holding_scope_for_offset(self, offset)

    @property
    @utils.saveit
    def _scope_finder(self):
        return _HoldingScopeFinder(self.pyobject)

    @property
    def builtin_names(self):
        return rope.base.builtins.builtins.get_attributes()


class ComprehensionScope(Scope):
    def __init__(self, pycore, pyobject, visitor):
        super().__init__(pycore, pyobject, pyobject.parent.get_scope())
        self.names = None
        self.returned_asts = None
        self.defineds = None
        self.visitor = visitor

    def _get_names(self):
        if self.names is None:
            self._visit_comprehension()
        return self.names

    def get_names(self):
        return self._get_names()

    def _visit_comprehension(self):
        if self.names is None:
            new_visitor = self.visitor(self.pycore, self.pyobject)
            for node in ast.iter_child_nodes(self.pyobject.get_ast()):
                new_visitor.visit(node)
            self.names = dict(self.parent.get_names())
            self.names.update(new_visitor.names)
            self.defineds = new_visitor.defineds

    def get_logical_end(self):
        return self.get_start()

    logical_end = property(get_logical_end)

    def get_body_start(self):
        return self.get_start()


class FunctionScope(Scope):
    def __init__(self, pycore, pyobject, visitor):
        super().__init__(pycore, pyobject, pyobject.parent.get_scope())
        self.names = None
        self.returned_asts = None
        self.is_generator = None
        self.defineds = None
        self.visitor = visitor

    def _get_names(self):
        if self.names is None:
            self._visit_function()
        return self.names

    def _visit_function(self):
        if self.names is None:
            new_visitor = self.visitor(self.pycore, self.pyobject)
            for n in ast.iter_child_nodes(self.pyobject.get_ast()):
                new_visitor.visit(n)
            self.names = new_visitor.names
            self.names.update(self.pyobject.get_parameters())
            self.returned_asts = new_visitor.returned_asts
            self.is_generator = new_visitor.generator
            self.defineds = new_visitor.defineds

    def _get_returned_asts(self):
        if self.names is None:
            self._visit_function()
        return self.returned_asts

    def _is_generator(self):
        if self.is_generator is None:
            self._get_returned_asts()
        return self.is_generator

    def get_names(self):
        return self._get_names()

    def _create_scopes(self):
        if self.defineds is None:
            self._visit_function()
        return [pydefined.get_scope() for pydefined in self.defineds]

    def get_kind(self):
        return "Function"

    def invalidate_data(self):
        for pyname in self.get_names().values():
            if isinstance(pyname, (pynames.AssignedName, pynames.EvaluatedName)):
                pyname.invalidate()


class ClassScope(Scope):
    def __init__(self, pycore, pyobject):
        super().__init__(pycore, pyobject, pyobject.parent.get_scope())

    def get_kind(self):
        return "Class"

    def get_propagated_names(self):
        return {}


class _HoldingScopeFinder:
    def __init__(self, pymodule):
        self.pymodule = pymodule

    def get_indents(self, lineno):
        return codeanalyze.count_line_indents(self.lines.get_line(lineno))

    def _get_scope_indents(self, scope):
        return self.get_indents(scope.get_start())

    def get_holding_scope(self, module_scope, lineno, line_indents=None):
        if line_indents is None:
            line_indents = self.get_indents(lineno)
        current_scope = module_scope
        new_scope = current_scope
        while new_scope is not None and (
            new_scope.get_kind() == "Module"
            or self._get_scope_indents(new_scope) <= line_indents
        ):
            current_scope = new_scope
            if (
                current_scope.get_start() == lineno
                and current_scope.get_kind() != "Module"
            ):
                return current_scope
            new_scope = None
            for scope in current_scope.get_scopes():
                if scope.get_start() <= lineno:
                    if lineno <= scope.get_end():
                        new_scope = scope
                        break
                else:
                    break
        return current_scope

    def _is_empty_line(self, lineno):
        line = self.lines.get_line(lineno)
        return line.strip() == "" or line.lstrip().startswith("#")

    def _get_body_indents(self, scope):
        return self.get_indents(scope.get_body_start())

    @staticmethod
    def get_holding_scope_for_offset(scope, offset):
        for inner_scope in scope.get_scopes():
            if inner_scope.in_region(offset):
                return _HoldingScopeFinder.get_holding_scope_for_offset(
                    inner_scope, offset
                )
        return scope

    def find_scope_end(self, scope):
        if not scope.parent:
            return self.lines.length()
        end = scope.pyobject.get_ast().body[-1].lineno
        scope_start = self.pymodule.logical_lines.logical_line_in(scope.start)
        if scope_start[1] >= end:
            # handling one-liners
            body_indents = self._get_scope_indents(scope) + 4
        else:
            body_indents = self._get_body_indents(scope)
        for line_start in self.logical_lines.generate_starts(
            min(end + 1, self.lines.length()), self.lines.length() + 1
        ):
            if not self._is_empty_line(line_start):
                if self.get_indents(line_start) < body_indents:
                    return end
                else:
                    end = line_start
        return end

    @property
    def lines(self):
        return self.pymodule.lines

    @property
    def code(self):
        return self.pymodule.source_code

    @property
    def logical_lines(self):
        return self.pymodule.logical_lines


class TemporaryScope(Scope):
    """Currently used for list comprehensions and generator expressions

    These scopes do not appear in the `get_scopes()` method of their
    parent scopes.
    """

    def __init__(self, pycore, parent_scope, names):
        super().__init__(pycore, parent_scope.pyobject, parent_scope)
        self.names = names

    def get_names(self):
        return self.names

    def get_defined_names(self):
        return self.names

    def _create_scopes(self):
        return []

    def get_kind(self):
        return "Temporary"



"""rope, a python refactoring library"""

from pkg_resources import DistributionNotFound, get_distribution

try:
    VERSION = get_distribution("rope").version
except DistributionNotFound:

    def get_fallback_version():
        import pathlib
        import re

        pyproject = (
            pathlib.Path(__file__).resolve().parent.parent / "pyproject.toml"
        ).read_text()
        version = re.search("version.*=.*'(.*)'", pyproject)
        return version.group(1) if version else None

    VERSION = get_fallback_version()


INFO = __doc__
COPYRIGHT = """\
Copyright (C) 2021-2022 Lie Ryan
Copyright (C) 2019-2021 Matej Cepl
Copyright (C) 2015-2018 Nicholas Smith
Copyright (C) 2014-2015 Matej Cepl
Copyright (C) 2006-2012 Ali Gholami Rudi
Copyright (C) 2009-2012 Anton Gritsay

This program is free software: you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this program.  If not, see
<https://www.gnu.org/licenses/>."""



import warnings

from rope.base import exceptions, resourceobserver
from rope.base.oi import memorydb, objectdb, transform


class ObjectInfoManager:
    """Stores object information

    It uses an instance of `objectdb.ObjectDB` for storing
    information.

    """

    def __init__(self, project):
        self.project = project
        self.to_textual = transform.PyObjectToTextual(project)
        self.to_pyobject = transform.TextualToPyObject(project)
        self.doi_to_pyobject = transform.DOITextualToPyObject(project)
        self._init_objectdb()
        if project.prefs.get("validate_objectdb", False):
            self._init_validation()

    def _init_objectdb(self):
        dbtype = self.project.get_prefs().get("objectdb_type", None)
        persist = None
        if dbtype is not None:
            warnings.warn(
                '"objectdb_type" project config is deprecated;\n'
                'Use "save_objectdb" instead in your project '
                'config file.\n(".ropeproject/config.py" by default)\n',
                DeprecationWarning,
            )
            if dbtype != "memory" and self.project.ropefolder is not None:
                persist = True
        self.validation = TextualValidation(self.to_pyobject)
        db = memorydb.MemoryDB(self.project, persist=persist)
        self.objectdb = objectdb.ObjectDB(db, self.validation)

    def _init_validation(self):
        self.objectdb.validate_files()
        observer = resourceobserver.ResourceObserver(
            changed=self._resource_changed,
            moved=self._resource_moved,
            removed=self._resource_moved,
        )
        files = []
        for path in self.objectdb.get_files():
            resource = self.to_pyobject.path_to_resource(path)
            if resource is not None and resource.project == self.project:
                files.append(resource)
        self.observer = resourceobserver.FilteredResourceObserver(observer, files)
        self.objectdb.add_file_list_observer(_FileListObserver(self))
        self.project.add_observer(self.observer)

    def _resource_changed(self, resource):
        try:
            self.objectdb.validate_file(self.to_textual.resource_to_path(resource))
        except exceptions.ModuleSyntaxError:
            pass

    def _resource_moved(self, resource, new_resource=None):
        self.observer.remove_resource(resource)
        if new_resource is not None:
            old = self.to_textual.resource_to_path(resource)
            new = self.to_textual.resource_to_path(new_resource)
            self.objectdb.file_moved(old, new)
            self.observer.add_resource(new_resource)

    def get_returned(self, pyobject, args):
        result = self.get_exact_returned(pyobject, args)
        if result is not None:
            return result
        path, key = self._get_scope(pyobject)
        if path is None:
            return None
        for call_info in self.objectdb.get_callinfos(path, key):
            returned = call_info.get_returned()
            if returned and returned[0] not in ("unknown", "none"):
                result = returned
                break
            if result is None:
                result = returned
        if result is not None:
            return self.to_pyobject(result)

    def get_exact_returned(self, pyobject, args):
        path, key = self._get_scope(pyobject)
        if path is not None:
            returned = self.objectdb.get_returned(
                path, key, self._args_to_textual(pyobject, args)
            )
            if returned is not None:
                return self.to_pyobject(returned)

    def _args_to_textual(self, pyfunction, args):
        parameters = list(pyfunction.get_param_names(special_args=False))
        arguments = args.get_arguments(parameters)[: len(parameters)]
        textual_args = tuple(self.to_textual(arg) for arg in arguments)
        return textual_args

    def get_parameter_objects(self, pyobject):
        path, key = self._get_scope(pyobject)
        if path is None:
            return None
        arg_count = len(pyobject.get_param_names(special_args=False))
        unknowns = arg_count
        parameters = [None] * arg_count
        for call_info in self.objectdb.get_callinfos(path, key):
            args = call_info.get_parameters()
            for index, arg in enumerate(args[:arg_count]):
                old = parameters[index]
                if self.validation.is_more_valid(arg, old):
                    parameters[index] = arg
                    if self.validation.is_value_valid(arg):
                        unknowns -= 1
            if unknowns == 0:
                break
        if unknowns < arg_count:
            return [self.to_pyobject(parameter) for parameter in parameters]

    def get_passed_objects(self, pyfunction, parameter_index):
        path, key = self._get_scope(pyfunction)
        if path is None:
            return []
        result = []
        for call_info in self.objectdb.get_callinfos(path, key):
            args = call_info.get_parameters()
            if len(args) > parameter_index:
                parameter = self.to_pyobject(args[parameter_index])
                if parameter is not None:
                    result.append(parameter)
        return result

    def doa_data_received(self, data):
        def doi_to_normal(textual):
            pyobject = self.doi_to_pyobject(textual)
            return self.to_textual(pyobject)

        function = doi_to_normal(data[0])
        args = tuple(doi_to_normal(textual) for textual in data[1])
        returned = doi_to_normal(data[2])
        if function[0] == "defined" and len(function) == 3:
            self._save_data(function, args, returned)

    def function_called(self, pyfunction, params, returned=None):
        function_text = self.to_textual(pyfunction)
        params_text = tuple(self.to_textual(param) for param in params)
        returned_text = ("unknown",)
        if returned is not None:
            returned_text = self.to_textual(returned)
        self._save_data(function_text, params_text, returned_text)

    def save_per_name(self, scope, name, data):
        path, key = self._get_scope(scope.pyobject)
        if path is not None:
            self.objectdb.add_pername(path, key, name, self.to_textual(data))

    def get_per_name(self, scope, name):
        path, key = self._get_scope(scope.pyobject)
        if path is not None:
            result = self.objectdb.get_pername(path, key, name)
            if result is not None:
                return self.to_pyobject(result)

    def _save_data(self, function, args, returned=("unknown",)):
        self.objectdb.add_callinfo(function[1], function[2], args, returned)

    def _get_scope(self, pyobject):
        resource = pyobject.get_module().get_resource()
        if resource is None:
            return None, None
        textual = self.to_textual(pyobject)
        if textual[0] == "defined":
            path = textual[1]
            if len(textual) == 3:
                key = textual[2]
            else:
                key = ""
            return path, key
        return None, None

    def sync(self):
        self.objectdb.sync()

    def __str__(self):
        return str(self.objectdb)


class TextualValidation:
    def __init__(self, to_pyobject):
        self.to_pyobject = to_pyobject

    def is_value_valid(self, value):
        # ???: Should none and unknown be considered valid?
        if value is None or value[0] in ("none", "unknown"):
            return False
        return self.to_pyobject(value) is not None

    def is_more_valid(self, new, old):
        if old is None:
            return True
        return new[0] not in ("unknown", "none")

    def is_file_valid(self, path):
        return self.to_pyobject.path_to_resource(path) is not None

    def is_scope_valid(self, path, key):
        if key == "":
            textual = ("defined", path)
        else:
            textual = ("defined", path, key)
        return self.to_pyobject(textual) is not None


class _FileListObserver:
    def __init__(self, object_info):
        self.object_info = object_info
        self.observer = self.object_info.observer
        self.to_pyobject = self.object_info.to_pyobject

    def removed(self, path):
        resource = self.to_pyobject.path_to_resource(path)
        if resource is not None:
            self.observer.remove_resource(resource)

    def added(self, path):
        resource = self.to_pyobject.path_to_resource(path)
        if resource is not None:
            self.observer.add_resource(resource)



import unittest
from textwrap import dedent

from rope.contrib import finderrors
from ropetest import testutils


class FindErrorsTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.mod = self.project.root.create_file("mod.py")

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_unresolved_variables(self):
        self.mod.write("print(var)\n")
        result = finderrors.find_errors(self.project, self.mod)
        self.assertEqual(1, len(result))
        self.assertEqual(1, result[0].lineno)

    def test_defined_later(self):
        self.mod.write(dedent("""\
            print(var)
            var = 1
        """))
        result = finderrors.find_errors(self.project, self.mod)
        self.assertEqual(1, len(result))
        self.assertEqual(1, result[0].lineno)

    def test_ignoring_builtins(self):
        self.mod.write("range(2)\n")
        result = finderrors.find_errors(self.project, self.mod)
        self.assertEqual(0, len(result))

    def test_ignoring_none(self):
        self.mod.write("var = None\n")
        result = finderrors.find_errors(self.project, self.mod)
        self.assertEqual(0, len(result))

    def test_bad_attributes(self):
        code = dedent("""\
            class C(object):
                pass
            c = C()
            print(c.var)
        """)
        self.mod.write(code)
        result = finderrors.find_errors(self.project, self.mod)
        self.assertEqual(1, len(result))
        self.assertEqual(4, result[0].lineno)



import unittest
from textwrap import dedent

from rope.base import libutils
from rope.base.pyobjects import get_base_type
from ropetest import testutils


class PyCoreScopesTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_simple_scope(self):
        code = dedent("""\
            def sample_func():
                pass
        """)
        scope = libutils.get_string_scope(self.project, code)

        sample_func = scope["sample_func"].get_object()
        self.assertEqual(get_base_type("Function"), sample_func.get_type())

    def test_simple_function_scope(self):
        code = dedent("""\
            def sample_func():
                a = 10
        """)
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(1, len(scope.get_scopes()))
        sample_func_scope = scope.get_scopes()[0]
        self.assertEqual(1, len(sample_func_scope.get_names()))
        self.assertEqual(0, len(sample_func_scope.get_scopes()))

    def test_classes_inside_function_scopes(self):
        code = dedent("""\
            def sample_func():
                class SampleClass(object):
                    pass
        """)
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(1, len(scope.get_scopes()))
        sample_func_scope = scope.get_scopes()[0]
        self.assertEqual(
            get_base_type("Type"),
            sample_func_scope["SampleClass"].get_object().get_type(),
        )

    def test_list_comprehension_scope_inside_assignment(self):
        code = "a_var = [b_var + d_var for b_var, c_var in e_var]\n"
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(
            list(sorted(scope.get_defined_names())),
            ["a_var"],
        )
        self.assertEqual(
            list(sorted(scope.get_scopes()[0].get_defined_names())),
            ["b_var", "c_var"],
        )

    def test_list_comprehension_scope(self):
        code = "[b_var + d_var for b_var, c_var in e_var]\n"
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(
            list(sorted(scope.get_scopes()[0].get_defined_names())),
            ["b_var", "c_var"],
        )

    def test_set_comprehension_scope(self):
        code = "{b_var + d_var for b_var, c_var in e_var}\n"
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(
            list(sorted(scope.get_scopes()[0].get_defined_names())),
            ["b_var", "c_var"],
        )

    def test_generator_comprehension_scope(self):
        code = "(b_var + d_var for b_var, c_var in e_var)\n"
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(
            list(sorted(scope.get_scopes()[0].get_defined_names())),
            ["b_var", "c_var"],
        )

    def test_dict_comprehension_scope(self):
        code = "{b_var: d_var for b_var, c_var in e_var}\n"
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(
            list(sorted(scope.get_scopes()[0].get_defined_names())),
            ["b_var", "c_var"],
        )

    @testutils.only_for_versions_higher("3.8")
    def test_inline_assignment(self):
        code = """values = (a_var := 2,)"""
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(
            list(sorted(scope.get_defined_names())),
            ["a_var", "values"],
        )

    @testutils.only_for_versions_higher("3.8")
    def test_inline_assignment_in_comprehensions(self):
        code = dedent("""\
            [
                (a_var := b_var + (f_var := g_var))
                for b_var in [(j_var := i_var)
                for i_var in c_var] if a_var + (h_var := d_var)
            ]
        """)
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(
            list(sorted(scope.get_scopes()[0].get_defined_names())),
            ["a_var", "b_var", "f_var"],
        )
        self.assertEqual(
            list(sorted(scope.get_scopes()[0].get_scopes()[0].get_defined_names())),
            ["i_var", "j_var"],
        )

    def test_nested_comprehension(self):
        code = dedent("""\
            [
                b_var + d_var for b_var, c_var in [
                    e_var for e_var in f_var
                ]
            ]
        """)
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(
            list(sorted(scope.get_scopes()[0].get_defined_names())),
            ["b_var", "c_var"],
        )
        self.assertEqual(
            list(sorted(scope.get_scopes()[0].get_scopes()[0].get_defined_names())),
            ["e_var"],
        )

    def test_simple_class_scope(self):
        code = dedent("""\
            class SampleClass(object):
                def f(self):
                    var = 10
        """)
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(1, len(scope.get_scopes()))
        sample_class_scope = scope.get_scopes()[0]
        self.assertTrue("f" in sample_class_scope)
        self.assertEqual(1, len(sample_class_scope.get_scopes()))
        f_in_class = sample_class_scope.get_scopes()[0]
        self.assertTrue("var" in f_in_class)

    def test_get_lineno(self):
        code = dedent("""\

            def sample_func():
                a = 10
        """)
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(1, len(scope.get_scopes()))
        sample_func_scope = scope.get_scopes()[0]
        self.assertEqual(1, scope.get_start())
        self.assertEqual(2, sample_func_scope.get_start())

    def test_scope_kind(self):
        code = dedent("""\
            class SampleClass(object):
                pass
            def sample_func():
                pass
        """)
        scope = libutils.get_string_scope(self.project, code)

        sample_class_scope = scope.get_scopes()[0]
        sample_func_scope = scope.get_scopes()[1]
        self.assertEqual("Module", scope.get_kind())
        self.assertEqual("Class", sample_class_scope.get_kind())
        self.assertEqual("Function", sample_func_scope.get_kind())

    def test_function_parameters_in_scope_names(self):
        code = dedent("""\
            def sample_func(param):
                a = 10
        """)
        scope = libutils.get_string_scope(self.project, code)

        sample_func_scope = scope.get_scopes()[0]
        self.assertTrue("param" in sample_func_scope)

    def test_get_names_contains_only_names_defined_in_a_scope(self):
        code = dedent("""\
            var1 = 10
            def sample_func(param):
                var2 = 20
        """)
        scope = libutils.get_string_scope(self.project, code)

        sample_func_scope = scope.get_scopes()[0]
        self.assertTrue("var1" not in sample_func_scope)

    def test_scope_lookup(self):
        code = dedent("""\
            var1 = 10
            def sample_func(param):
                var2 = 20
        """)
        scope = libutils.get_string_scope(self.project, code)

        self.assertTrue(scope.lookup("var2") is None)
        self.assertEqual(
            get_base_type("Function"),
            scope.lookup("sample_func").get_object().get_type(),
        )
        sample_func_scope = scope.get_scopes()[0]
        self.assertTrue(sample_func_scope.lookup("var1") is not None)

    def test_function_scopes(self):
        code = dedent("""\
            def func():
                var = 10
        """)
        scope = libutils.get_string_scope(self.project, code)

        func_scope = scope.get_scopes()[0]
        self.assertTrue("var" in func_scope)

    def test_function_scopes_classes(self):
        code = dedent("""\
            def func():
                class Sample(object):
                    pass
        """)
        scope = libutils.get_string_scope(self.project, code)

        func_scope = scope.get_scopes()[0]
        self.assertTrue("Sample" in func_scope)

    def test_function_getting_scope(self):
        code = dedent("""\
            def func():    var = 10
        """)
        mod = libutils.get_string_module(self.project, code)

        func_scope = mod["func"].get_object().get_scope()
        self.assertTrue("var" in func_scope)

    def test_scopes_in_function_scopes(self):
        code = dedent("""\
            def func():
                def inner():
                    var = 10
        """)
        scope = libutils.get_string_scope(self.project, code)

        func_scope = scope.get_scopes()[0]
        inner_scope = func_scope.get_scopes()[0]
        self.assertTrue("var" in inner_scope)

    def test_for_variables_in_scopes(self):
        code = dedent("""\
            for a_var in range(10):
                pass
        """)
        scope = libutils.get_string_scope(self.project, code)

        self.assertTrue("a_var" in scope)

    def test_assists_inside_fors(self):
        code = dedent("""\
            for i in range(10):
                a_var = i
        """)
        scope = libutils.get_string_scope(self.project, code)

        self.assertTrue("a_var" in scope)

    def test_first_parameter_of_a_method(self):
        code = dedent("""\
            class AClass(object):
                def a_func(self, param):
                    pass
        """)
        a_class = libutils.get_string_module(self.project, code)["AClass"].get_object()
        function_scope = a_class["a_func"].get_object().get_scope()
        self.assertEqual(a_class, function_scope["self"].get_object().get_type())
        self.assertNotEqual(a_class, function_scope["param"].get_object().get_type())

    def test_first_parameter_of_static_methods(self):
        code = dedent("""\
            class AClass(object):
                @staticmethod
                def a_func(param):
                    pass
        """)
        a_class = libutils.get_string_module(self.project, code)["AClass"].get_object()
        function_scope = a_class["a_func"].get_object().get_scope()
        self.assertNotEqual(a_class, function_scope["param"].get_object().get_type())

    def test_first_parameter_of_class_methods(self):
        code = dedent("""\
            class AClass(object):
                @classmethod
                def a_func(cls):
                    pass
        """)
        a_class = libutils.get_string_module(self.project, code)["AClass"].get_object()
        function_scope = a_class["a_func"].get_object().get_scope()
        self.assertEqual(a_class, function_scope["cls"].get_object())

    def test_first_parameter_with_self_as_name_and_unknown_decorator(self):
        code = dedent("""\
            def my_decorator(func):
                return func
            class AClass(object):
                @my_decorator
                def a_func(self):
                    pass
        """)
        a_class = libutils.get_string_module(self.project, code)["AClass"].get_object()
        function_scope = a_class["a_func"].get_object().get_scope()
        self.assertEqual(a_class, function_scope["self"].get_object().get_type())

    def test_inside_class_scope_attribute_lookup(self):
        code = dedent("""\
            class C(object):
                an_attr = 1
                def a_func(self):
                    pass""")
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(1, len(scope.get_scopes()))
        c_scope = scope.get_scopes()[0]
        self.assertTrue("an_attr" in c_scope.get_names())
        self.assertTrue(c_scope.lookup("an_attr") is not None)
        f_in_c = c_scope.get_scopes()[0]
        self.assertTrue(f_in_c.lookup("an_attr") is None)

    def test_inside_class_scope_attribute_lookup2(self):
        code = dedent("""\
            class C(object):
                def __init__(self):
                    self.an_attr = 1
                def a_func(self):
                    pass""")
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(1, len(scope.get_scopes()))
        c_scope = scope.get_scopes()[0]
        f_in_c = c_scope.get_scopes()[0]
        self.assertTrue(f_in_c.lookup("an_attr") is None)

    def test_get_inner_scope_for_staticmethods(self):
        code = dedent("""\
            class C(object):
                @staticmethod
                def a_func(self):
                    pass
        """)
        scope = libutils.get_string_scope(self.project, code)

        c_scope = scope.get_scopes()[0]
        f_in_c = c_scope.get_scopes()[0]
        self.assertEqual(f_in_c, scope.get_inner_scope_for_line(4))

    def test_get_scope_for_offset_for_comprehension(self):
        code = "a = [i for i in range(10)]\n"
        scope = libutils.get_string_scope(self.project, code)

        c_scope = scope.get_scopes()[0]
        self.assertEqual(c_scope, scope.get_inner_scope_for_offset(10))
        self.assertEqual(scope, scope.get_inner_scope_for_offset(1))

    def test_get_scope_for_offset_for_in_nested_comprehension(self):
        code = "[i for i in [j for j in k]]\n"
        scope = libutils.get_string_scope(self.project, code)

        c_scope = scope.get_scopes()[0]
        self.assertEqual(c_scope, scope.get_inner_scope_for_offset(5))
        inner_scope = c_scope.get_scopes()[0]
        self.assertEqual(inner_scope, scope.get_inner_scope_for_offset(15))

    def test_get_scope_for_offset_for_scope_with_indent(self):
        code = dedent("""\
            def f(a):
                print(a)
        """)
        scope = libutils.get_string_scope(self.project, code)

        inner_scope = scope.get_scopes()[0]
        self.assertEqual(inner_scope, scope.get_inner_scope_for_offset(10))

    def test_get_scope_for_offset_for_function_scope_and_async_with_statement(self):
        scope = libutils.get_string_scope(
            self.project,
            dedent("""\
                async def func():
                    async with a_func() as var:
                        print(var)
            """),
        )
        inner_scope = scope.get_scopes()[0]
        self.assertEqual(inner_scope, scope.get_inner_scope_for_offset(27))

    def test_getting_overwritten_scopes(self):
        code = dedent("""\
            def f():
                pass
            def f():
                pass
        """)
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(2, len(scope.get_scopes()))
        f1_scope = scope.get_scopes()[0]
        f2_scope = scope.get_scopes()[1]
        self.assertNotEqual(f1_scope, f2_scope)

    def test_assigning_builtin_names(self):
        code = "range = 1\n"
        mod = libutils.get_string_module(self.project, code)

        range = mod.get_scope().lookup("range")
        self.assertEqual((mod, 1), range.get_definition_location())

    def test_get_inner_scope_and_logical_lines(self):
        code = dedent('''\
            class C(object):
                def f():
                    s = """
            1
            2
            """
                    a = 1
        ''')
        scope = libutils.get_string_scope(self.project, code)

        c_scope = scope.get_scopes()[0]
        f_in_c = c_scope.get_scopes()[0]
        self.assertEqual(f_in_c, scope.get_inner_scope_for_line(7))

    def test_getting_defined_names_for_classes(self):
        code = dedent("""\
            class A(object):
                def a(self):
                    pass
            class B(A):
                def b(self):
                    pass
        """)
        scope = libutils.get_string_scope(self.project, code)

        a_scope = scope["A"].get_object().get_scope()  # noqa
        b_scope = scope["B"].get_object().get_scope()
        self.assertTrue("a" in b_scope.get_names())
        self.assertTrue("b" in b_scope.get_names())
        self.assertTrue("a" not in b_scope.get_defined_names())
        self.assertTrue("b" in b_scope.get_defined_names())

    def test_getting_defined_names_for_modules(self):
        code = dedent("""\
            class A(object):
                pass
        """)
        scope = libutils.get_string_scope(self.project, code)

        self.assertTrue("open" in scope.get_names())
        self.assertTrue("A" in scope.get_names())
        self.assertTrue("open" not in scope.get_defined_names())
        self.assertTrue("A" in scope.get_defined_names())

    def test_get_inner_scope_for_list_comprhension_with_many_targets(self):
        code = "a = [(i, j) for i,j in enumerate(range(10))]\n"
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(len(scope.get_scopes()), 1)
        self.assertNotIn("i", scope)
        self.assertNotIn("j", scope)
        self.assertIn("i", scope.get_scopes()[0])
        self.assertIn("j", scope.get_scopes()[0])

    def test_get_inner_scope_for_generator(self):
        code = "a = (i for i in range(10))\n"
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(len(scope.get_scopes()), 1)
        self.assertNotIn("i", scope)
        self.assertIn("i", scope.get_scopes()[0])

    def test_get_inner_scope_for_set_comprehension(self):
        code = "a = {i for i in range(10)}\n"
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(len(scope.get_scopes()), 1)
        self.assertNotIn("i", scope)
        self.assertIn("i", scope.get_scopes()[0])

    def test_get_inner_scope_for_dict_comprehension(self):
        code = "a = {i:i for i in range(10)}\n"
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(len(scope.get_scopes()), 1)
        self.assertNotIn("i", scope)
        self.assertIn("i", scope.get_scopes()[0])

    def test_get_inner_scope_for_nested_list_comprhension(self):
        code = "a = [[i + j for j in range(10)] for i in range(10)]\n"
        scope = libutils.get_string_scope(self.project, code)

        self.assertEqual(len(scope.get_scopes()), 1)
        self.assertNotIn("i", scope)
        self.assertNotIn("j", scope)
        self.assertIn("i", scope.get_scopes()[0])
        self.assertEqual(len(scope.get_scopes()[0].get_scopes()), 1)
        self.assertIn("j", scope.get_scopes()[0].get_scopes()[0])
        self.assertIn("i", scope.get_scopes()[0].get_scopes()[0])

    def test_get_scope_region(self):
        scope = libutils.get_string_scope(
            self.project,
            dedent("""
                def func1(ala):
                   pass

                def func2(o):
                   pass
            """),
        )

        self.assertEqual(scope.get_region(), (0, 48))
        self.assertEqual(scope.get_scopes()[0].get_region(), (1, 24))
        self.assertEqual(scope.get_scopes()[1].get_region(), (26, 47))

    def test_only_get_inner_scope_region(self):
        scope = libutils.get_string_scope(
            self.project,
            dedent("""
                def func1(ala):
                   pass

                def func2(o):
                   pass
            """),
        )

        self.assertEqual(scope.get_scopes()[1].get_region(), (26, 47))



import unittest
from textwrap import dedent

import rope.base.exceptions
from rope.refactor import inline
from ropetest import testutils


class InlineTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore
        self.mod = testutils.create_module(self.project, "mod")
        self.mod2 = testutils.create_module(self.project, "mod2")

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def _inline(self, code, offset, **kwds):
        self.mod.write(code)
        self._inline2(self.mod, offset, **kwds)
        return self.mod.read()

    def _inline2(self, resource, offset, **kwds):
        inliner = inline.create_inline(self.project, resource, offset)
        changes = inliner.get_changes(**kwds)
        self.project.do(changes)
        return self.mod.read()

    def test_simple_case(self):
        code = dedent("""\
            a_var = 10
            another_var = a_var
        """)
        refactored = self._inline(code, code.index("a_var") + 1)
        self.assertEqual("another_var = 10\n", refactored)

    def test_empty_case(self):
        code = "a_var = 10\n"
        refactored = self._inline(code, code.index("a_var") + 1)
        self.assertEqual("", refactored)

    def test_long_definition(self):
        code = dedent("""\
            a_var = 10 + (10 + 10)
            another_var = a_var
        """)
        refactored = self._inline(code, code.index("a_var") + 1)
        self.assertEqual("another_var = 10 + (10 + 10)\n", refactored)

    def test_explicit_continuation(self):
        code = dedent("""\
            a_var = (10 +
             10)
            another_var = a_var
        """)
        refactored = self._inline(code, code.index("a_var") + 1)
        self.assertEqual(
            dedent("""\
                another_var = (10 +
                 10)
            """),
            refactored,
        )

    def test_implicit_continuation(self):
        code = dedent("""\
            a_var = 10 +\\
                   10
            another_var = a_var
        """)
        refactored = self._inline(code, code.index("a_var") + 1)
        self.assertEqual(
            dedent("""\
                another_var = 10 +\\
                       10
            """),
            refactored,
        )

    def test_inlining_at_the_end_of_input(self):
        code = dedent("""\
            a = 1
            b = a""")
        refactored = self._inline(code, code.index("a") + 1)
        self.assertEqual("b = 1", refactored)

    def test_on_classes(self):
        code = dedent("""\
            class AClass(object):
                pass
        """)
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self._inline(code, code.index("AClass") + 1)

    def test_multiple_assignments(self):
        code = dedent("""\
            a_var = 10
            a_var = 20
        """)
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self._inline(code, code.index("a_var") + 1)

    def test_tuple_assignments(self):
        code = "a_var, another_var = (20, 30)\n"
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self._inline(code, code.index("a_var") + 1)

    def test_on_unknown_vars(self):
        code = "a_var = another_var\n"
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self._inline(code, code.index("another_var") + 1)

    def test_attribute_inlining(self):
        code = dedent("""\
            class A(object):
                def __init__(self):
                    self.an_attr = 3
                    range(self.an_attr)
        """)
        refactored = self._inline(code, code.index("an_attr") + 1)
        expected = dedent("""\
            class A(object):
                def __init__(self):
                    range(3)
        """)
        self.assertEqual(expected, refactored)

    def test_attribute_inlining2(self):
        code = dedent("""\
            class A(object):
                def __init__(self):
                    self.an_attr = 3
                    range(self.an_attr)
            a = A()
            range(a.an_attr)""")
        refactored = self._inline(code, code.index("an_attr") + 1)
        expected = dedent("""\
            class A(object):
                def __init__(self):
                    range(3)
            a = A()
            range(3)""")
        self.assertEqual(expected, refactored)

    def test_a_function_with_no_occurrence(self):
        self.mod.write(dedent("""\
            def a_func():
                pass
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual("", self.mod.read())

    def test_a_function_with_no_occurrence2(self):
        self.mod.write(dedent("""\
            a_var = 10
            def a_func():
                pass
            print(a_var)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                a_var = 10
                print(a_var)
            """),
            self.mod.read(),
        )

    def test_replacing_calls_with_function_definition_in_other_modules(self):
        self.mod.write(dedent("""\
            def a_func():
                print(1)
        """))
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            import mod
            mod.a_func()
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                import mod
                print(1)
            """),
            mod1.read(),
        )

    def test_replacing_calls_with_function_definition_in_other_modules2(self):
        self.mod.write(dedent("""\
            def a_func():
                print(1)
        """))
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            import mod
            if True:
                mod.a_func()
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                import mod
                if True:
                    print(1)
            """),
            mod1.read(),
        )

    def test_replacing_calls_with_method_definition_in_other_modules(self):
        self.mod.write(dedent("""\
            class A(object):
                var = 10
                def a_func(self):
                    print(1)
        """))
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            import mod
            mod.A().a_func()
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                import mod
                print(1)
            """),
            mod1.read(),
        )
        self.assertEqual(
            dedent("""\
                class A(object):
                    var = 10
            """),
            self.mod.read(),
        )

    def test_replacing_calls_with_function_definition_in_defining_module(self):
        self.mod.write(dedent("""\
            def a_func():
                print(1)
            a_func()
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual("print(1)\n", self.mod.read())

    def test_replac_calls_with_function_definition_in_defining_module2(self):
        self.mod.write(dedent("""\
            def a_func():
                for i in range(10):
                    print(1)
            a_func()
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                for i in range(10):
                    print(1)
            """),
            self.mod.read(),
        )

    def test_replacing_calls_with_method_definition_in_defining_modules(self):
        self.mod.write(dedent("""\
            class A(object):
                var = 10
                def a_func(self):
                    print(1)
            A().a_func()"""))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                class A(object):
                    var = 10
                print(1)
            """),
            self.mod.read(),
        )

    def test_parameters_with_the_same_name_as_passed(self):
        self.mod.write(dedent("""\
            def a_func(var):
                print(var)
            var = 1
            a_func(var)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                var = 1
                print(var)
            """),
            self.mod.read(),
        )

    def test_parameters_with_the_same_name_as_passed2(self):
        self.mod.write(dedent("""\
            def a_func(var):
                print(var)
            var = 1
            a_func(var=var)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                var = 1
                print(var)
            """),
            self.mod.read(),
        )

    def test_simple_parameters_renaming(self):
        self.mod.write(dedent("""\
            def a_func(param):
                print(param)
            var = 1
            a_func(var)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                var = 1
                print(var)
            """),
            self.mod.read(),
        )

    def test_simple_parameters_renaming_for_multiple_params(self):
        self.mod.write(dedent("""\
            def a_func(param1, param2):
                p = param1 + param2
            var1 = 1
            var2 = 1
            a_func(var1, var2)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                var1 = 1
                var2 = 1
                p = var1 + var2
            """),
            self.mod.read(),
        )

    def test_parameters_renaming_for_passed_constants(self):
        self.mod.write(dedent("""\
            def a_func(param):
                print(param)
            a_func(1)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual("print(1)\n", self.mod.read())

    def test_parameters_renaming_for_passed_statements(self):
        self.mod.write(dedent("""\
            def a_func(param):
                print(param)
            a_func((1 + 2) / 3)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                print((1 + 2) / 3)
            """),
            self.mod.read(),
        )

    def test_simple_parameters_renam_for_multiple_params_using_keywords(self):
        self.mod.write(dedent("""\
            def a_func(param1, param2):
                p = param1 + param2
            var1 = 1
            var2 = 1
            a_func(param2=var1, param1=var2)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                var1 = 1
                var2 = 1
                p = var2 + var1
            """),
            self.mod.read(),
        )

    def test_simple_params_renam_for_multi_params_using_mixed_keywords(self):
        self.mod.write(dedent("""\
            def a_func(param1, param2):
                p = param1 + param2
            var1 = 1
            var2 = 1
            a_func(var2, param2=var1)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                var1 = 1
                var2 = 1
                p = var2 + var1
            """),
            self.mod.read(),
        )

    def test_simple_putting_in_default_arguments(self):
        self.mod.write(dedent("""\
            def a_func(param=None):
                print(param)
            a_func()
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual("print(None)\n", self.mod.read())

    def test_overriding_default_arguments(self):
        self.mod.write(dedent("""\
            def a_func(param1=1, param2=2):
                print(param1, param2)
            a_func(param2=3)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual("print(1, 3)\n", self.mod.read())

    def test_arguments_containing_comparisons(self):
        self.mod.write(dedent("""\
            def a_func(param1, param2, param3):
                param2.name
            a_func(2 <= 1, item, True)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual("item.name\n", self.mod.read())

    def test_badly_formatted_text(self):
        self.mod.write(dedent("""\
            def a_func  (  param1 =  1 ,param2 = 2 )  :
                print(param1, param2)
            a_func  ( param2
                = 3 )
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual("print(1, 3)\n", self.mod.read())

    def test_passing_first_arguments_for_methods(self):
        a_class = dedent("""\
            class A(object):
                def __init__(self):
                    self.var = 1
                    self.a_func(self.var)
                def a_func(self, param):
                    print(param)
        """)
        self.mod.write(a_class)
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        expected = dedent("""\
            class A(object):
                def __init__(self):
                    self.var = 1
                    print(self.var)
        """)
        self.assertEqual(expected, self.mod.read())

    def test_passing_first_arguments_for_methods2(self):
        a_class = dedent("""\
            class A(object):
                def __init__(self):
                    self.var = 1
                def a_func(self, param):
                    print(param, self.var)
            an_a = A()
            an_a.a_func(1)
        """)
        self.mod.write(a_class)
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        expected = dedent("""\
            class A(object):
                def __init__(self):
                    self.var = 1
            an_a = A()
            print(1, an_a.var)
        """)
        self.assertEqual(expected, self.mod.read())

    def test_passing_first_arguments_for_methods3(self):
        a_class = dedent("""\
            class A(object):
                def __init__(self):
                    self.var = 1
                def a_func(self, param):
                    print(param, self.var)
            an_a = A()
            A.a_func(an_a, 1)
        """)
        self.mod.write(a_class)
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        expected = dedent("""\
            class A(object):
                def __init__(self):
                    self.var = 1
            an_a = A()
            print(1, an_a.var)
        """)
        self.assertEqual(expected, self.mod.read())

    def test_inlining_staticmethods(self):
        a_class = dedent("""\
            class A(object):
                @staticmethod
                def a_func(param):
                    print(param)
            A.a_func(1)
        """)
        self.mod.write(a_class)
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        expected = dedent("""\
            class A(object):
                pass
            print(1)
        """)
        self.assertEqual(expected, self.mod.read())

    def test_static_methods2(self):
        a_class = dedent("""\
            class A(object):
                var = 10
                @staticmethod
                def a_func(param):
                    print(param)
            an_a = A()
            an_a.a_func(1)
            A.a_func(2)
        """)
        self.mod.write(a_class)
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        expected = dedent("""\
            class A(object):
                var = 10
            an_a = A()
            print(1)
            print(2)
        """)
        self.assertEqual(expected, self.mod.read())

    def test_inlining_classmethods(self):
        a_class = dedent("""\
            class A(object):
                @classmethod
                def a_func(cls, param):
                    print(param)
            A.a_func(1)
        """)
        self.mod.write(a_class)
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        expected = dedent("""\
            class A(object):
                pass
            print(1)
        """)
        self.assertEqual(expected, self.mod.read())

    def test_inlining_classmethods2(self):
        a_class = dedent("""\
            class A(object):
                @classmethod
                def a_func(cls, param):
                    return cls
            print(A.a_func(1))
        """)
        self.mod.write(a_class)
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        expected = dedent("""\
            class A(object):
                pass
            print(A)
        """)
        self.assertEqual(expected, self.mod.read())

    def test_simple_return_values_and_inlining_functions(self):
        self.mod.write(dedent("""\
            def a_func():
                return 1
            a = a_func()
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual("a = 1\n", self.mod.read())

    def test_simple_return_values_and_inlining_lonely_functions(self):
        self.mod.write(dedent("""\
            def a_func():
                return 1
            a_func()
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual("1\n", self.mod.read())

    def test_empty_returns_and_inlining_lonely_functions(self):
        self.mod.write(dedent("""\
            def a_func():
                if True:
                    return
            a_func()
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                if True:
                    pass
            """),
            self.mod.read(),
        )

    def test_multiple_returns(self):
        self.mod.write(dedent("""\
            def less_than_five(var):
                if var < 5:
                    return True
                return False
            a = less_than_five(2)
        """))
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self._inline2(self.mod, self.mod.read().index("less") + 1)

    def test_multiple_returns_and_not_using_the_value(self):
        self.mod.write(dedent("""\
            def less_than_five(var):
                if var < 5:
                    return True
                return False
            less_than_five(2)
        """))
        self._inline2(self.mod, self.mod.read().index("less") + 1)
        self.assertEqual(
            dedent("""\
                if 2 < 5:
                    True
                False
            """),
            self.mod.read(),
        )

    def test_raising_exception_for_list_arguments(self):
        self.mod.write(dedent("""\
            def a_func(*args):
                print(args)
            a_func(1)
        """))
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self._inline2(self.mod, self.mod.read().index("a_func") + 1)

    def test_raising_exception_for_list_keywods(self):
        self.mod.write(dedent("""\
            def a_func(**kwds):
                print(kwds)
            a_func(n=1)
        """))
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self._inline2(self.mod, self.mod.read().index("a_func") + 1)

    def test_function_parameters_and_returns_in_other_functions(self):
        code = dedent("""\
            def a_func(param1, param2):
                return param1 + param2
            range(a_func(20, param2=abs(10)))
        """)
        self.mod.write(code)
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual("range(20 + abs(10))\n", self.mod.read())

    def test_function_references_other_than_call(self):
        self.mod.write(dedent("""\
            def a_func(param):
                print(param)
            f = a_func
        """))
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self._inline2(self.mod, self.mod.read().index("a_func") + 1)

    def test_function_referencing_itself(self):
        self.mod.write(dedent("""\
            def a_func(var):
                func = a_func
        """))
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self._inline2(self.mod, self.mod.read().index("a_func") + 1)

    def test_recursive_functions(self):
        self.mod.write(dedent("""\
            def a_func(var):
                a_func(var)
        """))
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self._inline2(self.mod, self.mod.read().index("a_func") + 1)

    # TODO: inlining on function parameters
    def xxx_test_inlining_function_default_parameters(self):
        self.mod.write(dedent("""\
            def a_func(p1=1):
                pass
            a_func()
        """))
        self._inline2(self.mod, self.mod.read().index("p1") + 1)
        self.assertEqual(
            dedent("""\
                def a_func(p1=1):
                    pass
                a_func()
            """),
            self.mod.read(),
        )

    def test_simple_inlining_after_extra_indented_lines(self):
        self.mod.write(dedent("""\
            def a_func():
                for i in range(10):
                    pass
            if True:
                pass
            a_func()
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                if True:
                    pass
                for i in range(10):
                    pass
            """),
            self.mod.read(),
        )

    def test_inlining_a_function_with_pydoc(self):
        self.mod.write(dedent('''\
            def a_func():
                """docs"""
                a = 1
            a_func()'''))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual("a = 1\n", self.mod.read())

    def test_inlining_methods(self):
        self.mod.write(dedent("""\
            class A(object):
                name = 'hey'
                def get_name(self):
                    return self.name
            a = A()
            name = a.get_name()
        """))
        self._inline2(self.mod, self.mod.read().rindex("get_name") + 1)
        self.assertEqual(
            dedent("""\
                class A(object):
                    name = 'hey'
                a = A()
                name = a.name
            """),
            self.mod.read(),
        )

    def test_simple_returns_with_backslashes(self):
        self.mod.write(dedent("""\
            def a_func():
                return 1\\
                    + 2
            a = a_func()
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            "a = 1\\\n    + 2\n",
            dedent("""\
                a = 1\\
                    + 2
            """),
            self.mod.read(),
        )

    def test_a_function_with_pass_body(self):
        self.mod.write(dedent("""\
            def a_func():
                print(1)
            a = a_func()
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                print(1)
                a = None
            """),
            self.mod.read(),
        )

    def test_inlining_the_last_method_of_a_class(self):
        self.mod.write(dedent("""\
            class A(object):
                def a_func(self):
                    pass
        """))
        self._inline2(self.mod, self.mod.read().rindex("a_func") + 1)
        self.assertEqual(
            dedent("""\
                class A(object):
                    pass
            """),
            self.mod.read(),
        )

    def test_adding_needed_imports_in_the_dest_module(self):
        self.mod.write(dedent("""\
            import sys

            def ver():
                print(sys.version)
        """))
        self.mod2.write(dedent("""\
            import mod

            mod.ver()"""))
        self._inline2(self.mod, self.mod.read().index("ver") + 1)
        self.assertEqual(
            dedent("""\
                import mod
                import sys

                print(sys.version)
            """),
            self.mod2.read(),
        )

    def test_adding_needed_imports_in_the_dest_module_removing_selfs(self):
        self.mod.write(dedent("""\
            import mod2

            def f():
                print(mod2.var)
        """))
        self.mod2.write(dedent("""\
            import mod

            var = 1
            mod.f()
        """))
        self._inline2(self.mod, self.mod.read().index("f(") + 1)
        self.assertEqual(
            dedent("""\
                import mod

                var = 1
                print(var)
            """),
            self.mod2.read(),
        )

    def test_handling_relative_imports_when_inlining(self):
        pkg = testutils.create_package(self.project, "pkg")
        mod3 = testutils.create_module(self.project, "mod3", pkg)
        mod4 = testutils.create_module(self.project, "mod4", pkg)
        mod4.write("var = 1\n")
        mod3.write(dedent("""\
            from . import mod4

            def f():
                print(mod4.var)
        """))
        self.mod.write(dedent("""\
            import pkg.mod3

            pkg.mod3.f()
        """))
        self._inline2(self.mod, self.mod.read().index("f(") + 1)
        # Cannot determine the exact import
        self.assertTrue("\n\nprint(mod4.var)\n" in self.mod.read())

    def test_adding_needed_imports_for_elements_in_source(self):
        self.mod.write(dedent("""\
            def f1():
                return f2()
            def f2():
                return 1
        """))
        self.mod2.write(dedent("""\
            import mod

            print(mod.f1())
        """))
        self._inline2(self.mod, self.mod.read().index("f1") + 1)
        self.assertEqual(
            dedent("""\
                import mod
                from mod import f2

                print(f2())
            """),
            self.mod2.read(),
        )

    def test_relative_imports_and_changing_inlining_body(self):
        pkg = testutils.create_package(self.project, "pkg")
        mod3 = testutils.create_module(self.project, "mod3", pkg)
        mod4 = testutils.create_module(self.project, "mod4", pkg)
        mod4.write("var = 1\n")
        mod3.write(dedent("""\
            import mod4

            def f():
                print(mod4.var)
        """))
        self.mod.write(dedent("""\
            import pkg.mod3

            pkg.mod3.f()
        """))
        self._inline2(self.mod, self.mod.read().index("f(") + 1)
        self.assertEqual(
            dedent("""\
                import pkg.mod3
                import pkg.mod4

                print(pkg.mod4.var)
            """),
            self.mod.read(),
        )

    def test_inlining_with_different_returns(self):
        self.mod.write(dedent("""\
            def f(p):
                return p
            print(f(1))
            print(f(2))
            print(f(1))
        """))
        self._inline2(self.mod, self.mod.read().index("f(") + 1)
        self.assertEqual(
            dedent("""\
                print(1)
                print(2)
                print(1)
            """),
            self.mod.read(),
        )

    def test_not_removing_definition_for_variables(self):
        code = dedent("""\
            a_var = 10
            another_var = a_var
        """)
        refactored = self._inline(code, code.index("a_var") + 1, remove=False)
        self.assertEqual(
            dedent("""\
                a_var = 10
                another_var = 10
            """),
            refactored,
        )

    def test_not_removing_definition_for_methods(self):
        code = dedent("""\
            def func():
                print(1)

            func()
        """)
        refactored = self._inline(code, code.index("func") + 1, remove=False)
        self.assertEqual(
            dedent("""\
                def func():
                    print(1)

                print(1)
            """),
            refactored,
        )

    def test_only_current_for_methods(self):
        code = dedent("""\
            def func():
                print(1)

            func()
            func()
        """)
        refactored = self._inline(
            code, code.rindex("func") + 1, remove=False, only_current=True
        )
        self.assertEqual(
            dedent("""\
                def func():
                    print(1)

                func()
                print(1)
            """),
            refactored,
        )

    def test_only_current_for_variables(self):
        code = dedent("""\
            one = 1

            a = one
            b = one
        """)
        refactored = self._inline(
            code, code.rindex("one") + 1, remove=False, only_current=True
        )
        self.assertEqual(
            dedent("""\
                one = 1

                a = one
                b = 1
            """),
            refactored,
        )

    def test_inlining_one_line_functions(self):
        code = dedent("""\
            def f(): return 1
            var = f()
        """)
        refactored = self._inline(code, code.rindex("f"))
        self.assertEqual("var = 1\n", refactored)

    def test_inlining_one_line_functions_with_breaks(self):
        code = dedent("""\
            def f(
            p): return p
            var = f(1)
        """)
        refactored = self._inline(code, code.rindex("f"))
        self.assertEqual("var = 1\n", refactored)

    def test_inlining_one_line_functions_with_breaks2(self):
        code = dedent("""\
            def f(
            ): return 1
            var = f()
        """)
        refactored = self._inline(code, code.rindex("f"))
        self.assertEqual("var = 1\n", refactored)

    def test_resources_parameter(self):
        self.mod.write(dedent("""\
            def a_func():
                print(1)
        """))
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            import mod
            mod.a_func()
        """))
        self._inline2(self.mod, self.mod.read().index("a_func"), resources=[self.mod])
        self.assertEqual("", self.mod.read())
        self.assertEqual(
            dedent("""\
                import mod
                mod.a_func()
            """),
            mod1.read(),
        )

    def test_inlining_parameters(self):
        code = dedent("""\
            def f(p=1):
                pass
            f()
        """)
        result = self._inline(code, code.index("p"))
        self.assertEqual(
            dedent("""\
                def f(p=1):
                    pass
                f(1)
            """),
            result,
        )

    def test_inlining_function_with_line_breaks_in_args(self):
        code = dedent("""\
            def f(p): return p
            var = f(1 +
            1)
        """)
        refactored = self._inline(code, code.rindex("f"))
        self.assertEqual("var = 1 + 1\n", refactored)

    def test_inlining_variables_before_comparison(self):
        code = "start = 1\nprint(start <= 2)\n"
        refactored = self._inline(code, code.index("start"))
        self.assertEqual("print(1 <= 2)\n", refactored)

    def test_inlining_variables_in_other_modules(self):
        self.mod.write("myvar = 1\n")
        self.mod2.write(dedent("""\
            import mod
            print(mod.myvar)
        """))
        self._inline2(self.mod, 2)
        self.assertEqual(
            dedent("""\
                import mod
                print(1)
            """),
            self.mod2.read(),
        )

    def test_inlining_variables_and_back_importing(self):
        self.mod.write(dedent("""\
            mainvar = 1
            myvar = mainvar
        """))
        self.mod2.write(dedent("""\
            import mod
            print(mod.myvar)
        """))
        self._inline2(self.mod, self.mod.read().index("myvar"))
        expected = dedent("""\
            import mod
            from mod import mainvar
            print(mainvar)
        """)
        self.assertEqual(expected, self.mod2.read())

    def test_inlining_variables_and_importing_used_imports(self):
        self.mod.write(dedent("""\
            import sys
            myvar = sys.argv
        """))
        self.mod2.write(dedent("""\
            import mod
            print(mod.myvar)
        """))
        self._inline2(self.mod, self.mod.read().index("myvar"))
        expected = dedent("""\
            import mod
            import sys
            print(sys.argv)
        """)
        self.assertEqual(expected, self.mod2.read())

    def test_inlining_variables_and_removing_old_froms(self):
        self.mod.write("var = 1\n")
        self.mod2.write(dedent("""\
            from mod import var
            print(var)
        """))
        self._inline2(self.mod2, self.mod2.read().rindex("var"))
        self.assertEqual("print(1)\n", self.mod2.read())

    def test_inlining_method_and_removing_old_froms(self):
        self.mod.write(dedent("""\
            def f():    return 1
        """))
        self.mod2.write(dedent("""\
            from mod import f
            print(f())
        """))
        self._inline2(self.mod2, self.mod2.read().rindex("f"))
        self.assertEqual("print(1)\n", self.mod2.read())

    def test_inlining_functions_in_other_modules_and_only_current(self):
        code1 = dedent("""\
            def f():
                return 1
            print(f())
        """)
        code2 = dedent("""\
            import mod
            print(mod.f())
            print(mod.f())
        """)
        self.mod.write(code1)
        self.mod2.write(code2)
        self._inline2(
            self.mod2, self.mod2.read().rindex("f"), remove=False, only_current=True
        )
        expected2 = dedent("""\
            import mod
            print(mod.f())
            print(1)
        """)
        self.assertEqual(code1, self.mod.read())
        self.assertEqual(expected2, self.mod2.read())

    def test_inlining_variables_in_other_modules_and_only_current(self):
        code1 = dedent("""\
            var = 1
            print(var)
        """)
        code2 = dedent("""\
            import mod
            print(mod.var)
            print(mod.var)
        """)
        self.mod.write(code1)
        self.mod2.write(code2)
        self._inline2(
            self.mod2, self.mod2.read().rindex("var"), remove=False, only_current=True
        )
        expected2 = "import mod\n" "print(mod.var)\n" "print(1)\n"
        self.assertEqual(code1, self.mod.read())
        self.assertEqual(expected2, self.mod2.read())

    def test_inlining_does_not_change_string_constants(self):
        code = dedent("""\
            var = 1
            print("var\\
            ")
        """)
        expected = dedent("""\
            var = 1
            print("var\\
            ")
        """)
        refactored = self._inline(
            code, code.rindex("var"), remove=False, only_current=True, docs=False
        )
        self.assertEqual(expected, refactored)

    def test_inlining_does_change_string_constants_if_docs_is_set(self):
        code = dedent("""\
            var = 1
            print("var\\
            ")
        """)
        expected = dedent("""\
            var = 1
            print("1\\
            ")
        """)
        refactored = self._inline(
            code, code.rindex("var"), remove=False, only_current=True, docs=True
        )
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.6")
    def test_inlining_into_format_string(self):
        code = dedent("""\
            var = 123
            print(f"{var}")
        """)
        expected = dedent("""\
            print(f"{123}")
        """)

        refactored = self._inline(code, code.rindex("var"))

        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.6")
    def test_inlining_into_format_string_containing_quotes(self):
        code = dedent('''\
            var = 123
            print(f" '{var}' ")
            print(f""" "{var}" """)
            print(f' "{var}" ')
        ''')
        expected = dedent('''\
            print(f" '{123}' ")
            print(f""" "{123}" """)
            print(f' "{123}" ')
        ''')

        refactored = self._inline(code, code.rindex("var"))

        self.assertEqual(expected, refactored)

    def test_parameters_with_the_same_name_as_passed_with_type_hints(self):
        self.mod.write(dedent("""\
            def a_func(var: int):
                print(var)
            var = 1
            a_func(var)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                var = 1
                print(var)
            """),
            self.mod.read(),
        )

    def test_parameters_with_the_same_name_as_passed_as_kwargs_with_type_hints(self):
        self.mod.write(dedent("""\
            def a_func(var: int):
                print(var)
            var = 1
            a_func(var=var)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                var = 1
                print(var)
            """),
            self.mod.read(),
        )

    def test_simple_parameters_renaming_with_type_hints(self):
        self.mod.write(dedent("""\
            def a_func(param: int):
                print(param)
            var = 1
            a_func(var)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                var = 1
                print(var)
            """),
            self.mod.read(),
        )

    def test_simple_parameters_renaming_for_multiple_params_with_type_hints(self):
        self.mod.write(dedent("""\
            def a_func(param1, param2: int):
                p = param1 + param2
            var1 = 1
            var2 = 1
            a_func(var1, var2)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                var1 = 1
                var2 = 1
                p = var1 + var2
            """),
            self.mod.read(),
        )

    def test_parameters_renaming_for_passed_constants_with_type_hints(self):
        self.mod.write(dedent("""\
            def a_func(param: int):
                print(param)
            a_func(1)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual("print(1)\n", self.mod.read())

    def test_parameters_renaming_for_passed_statements_with_type_hints(self):
        self.mod.write(dedent("""\
            def a_func(param: int):
                print(param)
            a_func((1 + 2) / 3)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                print((1 + 2) / 3)
            """),
            self.mod.read(),
        )

    def test_simple_parameters_renaming_for_multiple_params_using_keywords_with_type_hints(
        self,
    ):
        self.mod.write(dedent("""\
            def a_func(param1, param2: int):
                p = param1 + param2
            var1 = 1
            var2 = 1
            a_func(param2=var1, param1=var2)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                var1 = 1
                var2 = 1
                p = var2 + var1
            """),
            self.mod.read(),
        )

    def test_simple_params_renaming_for_multi_params_using_mixed_keywords_with_type_hints(
        self,
    ):
        self.mod.write(dedent("""\
            def a_func(param1, param2: int):
                p = param1 + param2
            var1 = 1
            var2 = 1
            a_func(var2, param2=var1)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual(
            dedent("""\
                var1 = 1
                var2 = 1
                p = var2 + var1
            """),
            self.mod.read(),
        )

    def test_simple_putting_in_default_arguments_with_type_hints(self):
        self.mod.write(dedent("""\
            def a_func(param: Optional[int] = None):
                print(param)
            a_func()
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual("print(None)\n", self.mod.read())

    def test_overriding_default_arguments_with_type_hints(self):
        self.mod.write(dedent("""\
            def a_func(param1=1, param2: int = 2):
                print(param1, param2)
            a_func(param2=3)
        """))
        self._inline2(self.mod, self.mod.read().index("a_func") + 1)
        self.assertEqual("print(1, 3)\n", self.mod.read())

    def test_dictionary_with_inline_comment(self):
        code = dedent("""\
            myvar = {
                "key": "value",  # noqa
            }
            print(myvar)
        """)
        refactored = self._inline(code, code.index("myvar") + 1)
        expected = dedent("""\
            print({
                "key": "value",  # noqa
            })
        """)
        self.assertEqual(expected, refactored)
